#Imports
import time
import math
import momepy
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import json
import random
from momepy.datasets import get_path
import geopandas as gpd
import networkx as nx
from shapely.geometry import Point

# Show maps inline
import bokeh.io
from bokeh.resources import INLINE
bokeh.io.output_notebook(INLINE)
from bokeh.plotting import figure, show
from bokeh.models import HoverTool, Circle, MultiLine
from bokeh.tile_providers import CARTODBPOSITRON
from pyproj import Transformer
from bokeh.models import HoverTool, ColumnDataSource, ColorBar, LogColorMapper
from bokeh.tile_providers import get_provider, Vendors
from bokeh.plotting import figure, show
from bokeh.palettes import RdYlBu
from bokeh.palettes import Viridis256

def merge_duplicate_nodes(G, node_attribute='is_station', secondary_attribute='AREA', third_attribute='fid', fourth_attribute='ACC_TIME', fifth_attribute='NAME'):
    """
    Merges nodes in the graph G that have the same coordinates and prioritises the attributes
    of the node where 'is_station' is set to True, so that attributes from the station node
    are preserved in the merged node. Also preserves edges with a focus on the 'AREA' attribute.

    :param G: NetworkX graph
    :param node_attribute: the primary attribute to check for ('is_station')
    :param secondary_attribute: an additional attribute to merge ('AREA')
    :param third_attribute: an additional attribute to merge ('fid')
    :param fourth_attribute: an additional attribute to merge ('ACC_TIME')
    :param fifth_attribute: an additional attribute to merge ('NAME')
    :return: A new graph with merged nodes and preserved edge attributes
    """
    new_graph = nx.Graph()
    coord_to_new_node = {}

    # Merge nodes based on exact coordinates
    for node, data in G.nodes(data=True):
        coord = (data['lat'], data['lon'])
        if coord in coord_to_new_node:
            existing_node = coord_to_new_node[coord]
            existing_data = new_graph.nodes[existing_node]

            # Check if either node is a station and prioritise that data
            if data.get(node_attribute) or existing_data.get(node_attribute):
                preferred_data = data if data.get(node_attribute) else existing_data
                new_graph.nodes[existing_node].update({
                    node_attribute: True,  # Since one of them is a station
                    secondary_attribute: preferred_data.get(secondary_attribute),
                    third_attribute: preferred_data.get(third_attribute),
                    fourth_attribute: preferred_data.get(fourth_attribute),
                    fifth_attribute: preferred_data.get(fifth_attribute)
                })
        else:
            new_node = len(new_graph)
            new_graph.add_node(new_node, **data)
            coord_to_new_node[coord] = new_node

    # Merge edges with priority to 'AREA' attribute
    edge_data_by_pair = {}
    for n1, n2, data in G.edges(data=True):
        new_n1 = coord_to_new_node[(G.nodes[n1]['lat'], G.nodes[n1]['lon'])]
        new_n2 = coord_to_new_node[(G.nodes[n2]['lat'], G.nodes[n2]['lon'])]
        if (new_n1, new_n2) not in edge_data_by_pair and (new_n2, new_n1) not in edge_data_by_pair:
            edge_data_by_pair[(new_n1, new_n2)] = data
        else:
            # Resolve conflicts where edges might already exist
            if (new_n2, new_n1) in edge_data_by_pair:
                existing_data = edge_data_by_pair[(new_n2, new_n1)]
            else:
                existing_data = edge_data_by_pair[(new_n1, new_n2)]

            # Prioritise edges where 'AREA' is not None
            if existing_data['AREA'] is None and data['AREA'] is not None:
                edge_data_by_pair[(new_n1, new_n2)] = data
            elif existing_data['AREA'] is not None and data['AREA'] is None:
                continue  # Keep existing
            else:
                # Prefer the edge with the lower weight if 'AREA' is the same
                if data['weight'] < existing_data['weight']:
                    edge_data_by_pair[(new_n1, new_n2)] = data

    # Add the determined best edges to the new graph
    for (n1, n2), data in edge_data_by_pair.items():
        new_graph.add_edge(n1, n2, **data)

    return new_graph

def merge_close_nodes(G, decimal_precision):
    """
    Merges nodes in the graph G that are close to each other (within a threshold based on decimal precision),
    prioritizing the attributes of station nodes ('is_station': True) and preserving edges with non-null 'AREA' attributes during merging.

    :param G: NetworkX graph
    :param decimal_precision: the number of decimal places to consider for the coordinates, affecting the merge threshold
    :return: A new graph with merged nodes
    """
    new_graph = nx.Graph()

    coord_to_new_node = {}

    # Merge nodes based on proximity
    for node, data in G.nodes(data=True):
        # Round the coordinates to the specified precision
        rounded_coord = (round(data['lat'], decimal_precision), round(data['lon'], decimal_precision))

        if rounded_coord in coord_to_new_node:
            existing_node = coord_to_new_node[rounded_coord]
            existing_data = new_graph.nodes[existing_node]

            # Prioritise station data
            preferred_data = data if data.get('is_station', False) else existing_data
            if data.get('is_station', False) or existing_data.get('is_station', False):
                preferred_data['is_station'] = True

            new_graph.nodes[existing_node].update(preferred_data)
        else:
            new_node = len(new_graph)
            new_graph.add_node(new_node, **data)
            coord_to_new_node[rounded_coord] = new_node

    # Merge edges with priority to 'AREA' attribute
    edge_data_by_pair = {}
    for n1, n2, data in G.edges(data=True):
        new_n1 = coord_to_new_node[(round(G.nodes[n1]['lat'], decimal_precision), round(G.nodes[n1]['lon'], decimal_precision))]
        new_n2 = coord_to_new_node[(round(G.nodes[n2]['lat'], decimal_precision), round(G.nodes[n2]['lon'], decimal_precision))]
        if (new_n1, new_n2) not in edge_data_by_pair and (new_n2, new_n1) not in edge_data_by_pair:
            edge_data_by_pair[(new_n1, new_n2)] = data
        else:
            # Correct syntax for tuple checking in dictionary keys
            if (new_n2, new_n1) in edge_data_by_pair:
                existing_data = edge_data_by_pair[(new_n2, new_n1)]
            else:
                existing_data = edge_data_by_pair[(new_n1, new_n2)]

            # Check if existing edge data has 'AREA' as None and new data has 'AREA' not None
            if existing_data['AREA'] is None and data['AREA'] is not None:
                edge_data_by_pair[(new_n1, new_n2)] = data
            elif existing_data['AREA'] is not None and data['AREA'] is None:
                continue  # Keep existing
            else:
                # If both are non-None or both None, prefer the one with the lower weight (or other criteria)
                if data['weight'] < existing_data['weight']:
                    edge_data_by_pair[(new_n1, new_n2)] = data

    # Add the determined best edges to the new graph
    for (n1, n2), data in edge_data_by_pair.items():
        new_graph.add_edge(n1, n2, **data)

    return new_graph

# Function to count the number of station nodes in the graph
def count_station_nodes(G):
    station_nodes = [node for node, data in G.nodes(data=True) if data.get('is_station') == True]
    return len(station_nodes)

def sanity_check(G):
    print("Checking self loops...")
    count=0
    for n in G.edges:
        if n[0]==n[1]:
            print("Self loop found: %d. Consider removing it manually."%n[0])
            count+=1
    print("Number of self-loops", count)
    print("---")

    print("Number of edges: ", len(G.edges()))
    print("Number of nodes: ", len(G.nodes()))
   # print("Number of strongly connected components: %d"%nx.number_strongly_connected_components(G))

def remove_self_loops(G):
    self_loops = list(nx.selfloop_edges(G))
    count = len(self_loops)
    if count > 0:
        G.remove_edges_from(self_loops)
        print("Removing of self-loops is done.")
    else:
        print("No self loops found.")
    print("---")


def lat_lon_to_web_mercator(lat, lon):
    k = 6378137
    x = lon * (k * np.pi/180.0)
    y = np.log(np.tan((90 + lat) * np.pi/360.0)) * k
    return x, y

def plot_graph_with_shortest_path(G, shortest_paths, start_node, end_node):
    pos = {node: lat_lon_to_web_mercator(data['lat'], data['lon']) for node, data in G.nodes(data=True)}

    # Define figure
    p = figure(height=600, width=950,
               x_range=(min([x for x, y in pos.values()]), max([x for x, y in pos.values()])),
               y_range=(min([y for x, y in pos.values()]), max([y for x, y in pos.values()])),
               x_axis_type="mercator", y_axis_type="mercator")
    
    # Add the CARTODBPOSITRON tile source for the map background
    p.add_tile(CARTODBPOSITRON)

    # Plot all edges
    edge_xs, edge_ys = [], []
    for edge in G.edges():
        x0, y0 = pos[edge[0]]
        x1, y1 = pos[edge[1]]
        edge_xs.append([x0, x1])
        edge_ys.append([y0, y1])
    p.multi_line(edge_xs, edge_ys, color="#888888", line_width=1)

    # Highlight the shortest path in blue
    if (start_node, end_node) in shortest_paths:
        path = shortest_paths[(start_node, end_node)]['path']
        path_xs, path_ys = [], []
        for i in range(len(path)-1):
            x0, y0 = pos[path[i]]
            x1, y1 = pos[path[i+1]]
            path_xs.append([x0, x1])
            path_ys.append([y0, y1])
        p.multi_line(path_xs, path_ys, color="blue", line_width=3)

    # Prepare node data for plotting and hover information
    station_xs, station_ys, lats, lons, areas, fids, acc_times, names = [], [], [], [], [], [], [], []
    for node, (x, y) in pos.items():
        data = G.nodes[node]
        if data.get('is_station', False):  # Check if node is a station
            station_xs.append(x)
            station_ys.append(y)
            lats.append(data['lat'])
            lons.append(data['lon'])
            areas.append(data['AREA'])
            fids.append(data['fid'])
            acc_times.append(data['ACC_TIME'])
            names.append(data['NAME'])

    # Create a ColumnDataSource for station nodes
    source_stations = ColumnDataSource({
        'x': station_xs,
        'y': station_ys,
        'lat': lats,
        'lon': lons,
        'AREA': areas,
        'fid': fids,
        'ACC_TIME': acc_times,
        'NAME': names
    })

    # Plot station nodes
    p.circle('x', 'y', source=source_stations, size=10, color="#fc8d59", fill_alpha=0.8)

    # Add hover tool for nodes
    hover_nodes = HoverTool(renderers=[p.circle('x', 'y', source=source_stations)], tooltips=[
        ("Latitude", "@lat"),
        ("Longitude", "@lon"),
        ("Area", "@AREA"),
        ("FID", "@fid"),
        ("Access Time", "@ACC_TIME"),
        ("Name", "@NAME")
    ])
    p.add_tools(hover_nodes)

    # Show the plot
    show(p)

def check_bidirectionality(od_pairs):
    count=0
    for key in od_pairs:
        origin, destination = key
        reverse_key = (destination, origin)
        if reverse_key in od_pairs:
            print(f"Path between {key} is bidirectional.")
            count+=1
            # You can add more actions here if needed
    if count == 0:
        print('All paths are unidirectional')

# Define a function to serialise the path for uniqueness comparison
def serialize_path(path):
    return "-".join(map(str, path))

def calculate_path_size_factors(od_pairs, graph):
    psfs = {}
    for od_pair, paths in od_pairs.items():
        path_edges_list = []  # To store edges of all paths for this OD pair
        total_costs = []  # To store total cost of each path

        # First, we need to gather all edges for each path and their total costs
        for path in paths:
            path_edges = [(path['path'][i], path['path'][i+1]) for i in range(len(path['path']) - 1)]
            total_cost = sum(graph[path_edges[i][0]][path_edges[i][1]]['weight'] for i in range(len(path_edges)))
            path_edges_list.append(path_edges)
            total_costs.append(total_cost)

        # Now calculate PSF for each path
        psfs[od_pair] = []
        for path_idx, path_edges in enumerate(path_edges_list):
            path_size_factor = 0
            for edge in path_edges:
                # Count the number of paths that include this edge
                edge_count = sum(edge in pe for pe in path_edges_list)
                edge_weight = graph[edge[0]][edge[1]]['weight']
                # Calculate the contribution of this edge to the path size factor
                path_size_factor += edge_weight / (total_costs[path_idx] * edge_count)
            psfs[od_pair].append(path_size_factor)

    return psfs

def calculate_psl_probabilities(od_pairs, psfs):
    for od_pair, paths in od_pairs.items():
        # Retrieve the path size factors for this OD pair
        path_size_factors = psfs.get(od_pair, [])
        # Calculate the numerator of the PSL probability for each path
        exp_values = [
            math.exp(path['utility'] + math.log(psf)) 
            for path, psf in zip(paths, path_size_factors)
        ]
        # Calculate the sum of all exponentiated utilities for normalization
        sum_exp_values = sum(exp_values)
        # Assign probabilities to each path
        for path, exp_value in zip(paths, exp_values):
            path['probability'] = exp_value / sum_exp_values

def plot_graph_with_multiple_paths(G, od_pairs, og_area, dest_area):
    pos = {node: lat_lon_to_web_mercator(data['lat'], data['lon']) for node, data in G.nodes(data=True)}

    # Define figure
    p = figure(height=600, width=950,
               x_range=(min([x for x, y in pos.values()]), max([x for x, y in pos.values()])),
               y_range=(min([y for x, y in pos.values()]), max([y for x, y in pos.values()])),
               x_axis_type="mercator", y_axis_type="mercator")

    # Add the CARTODBPOSITRON tile source for the map background
    p.add_tile(CARTODBPOSITRON)

    # Plot all edges
    edge_xs, edge_ys = [], []
    for edge in G.edges():
        x0, y0 = pos[edge[0]]
        x1, y1 = pos[edge[1]]
        edge_xs.append([x0, x1])
        edge_ys.append([y0, y1])
    p.multi_line(edge_xs, edge_ys, color="#888888", line_width=1)

    # Prepare colors for multiple paths
    colors = ['#fc8d59', '#91bfdb', '#99d594', '#fee08b', '#d73027']  # Define more colors as needed

    # Plot paths for the specific OD-pair
    if (og_area, dest_area) in od_pairs:
        paths = od_pairs[(og_area, dest_area)]
        for idx, path_details in enumerate(paths):
            path = path_details['path']
            path_xs, path_ys = [], []
            for i in range(len(path)-1):
                x0, y0 = pos[path[i]]
                x1, y1 = pos[path[i+1]]
                path_xs.append([x0, x1])
                path_ys.append([y0, y1])
            p.multi_line(path_xs, path_ys, color=colors[idx % len(colors)], line_width=3, legend_label=f'Path {idx+1}')

    # Prepare node data for plotting and hover information
    station_xs, station_ys, lats, lons, areas, fids, acc_times, names = [], [], [], [], [], [], [], []
    for node, (x, y) in pos.items():
        data = G.nodes[node]
        if data.get('is_station', False):  # Check if node is a station
            station_xs.append(x)
            station_ys.append(y)
            lats.append(data['lat'])
            lons.append(data['lon'])
            areas.append(data.get('AREA', 'N/A'))  # Handle missing data
            fids.append(data.get('fid', 'N/A'))
            acc_times.append(data.get('ACC_TIME', 'N/A'))
            names.append(data.get('NAME', 'N/A'))

    # Create a ColumnDataSource for station nodes
    source_stations = ColumnDataSource({
        'x': station_xs,
        'y': station_ys,
        'lat': lats,
        'lon': lons,
        'AREA': areas,
        'fid': fids,
        'ACC_TIME': acc_times,
        'NAME': names
    })

    # Plot station nodes
    p.circle('x', 'y', source=source_stations, size=10, color="#fc8d59", fill_alpha=0.8)

    # Add hover tool for nodes
    hover_nodes = HoverTool(renderers=[p.circle('x', 'y', source=source_stations)], tooltips=[
        ("Latitude", "@lat"),
        ("Longitude", "@lon"),
        ("Area", "@AREA"),
        ("FID", "@fid"),
        ("Access Time", "@ACC_TIME"),
        ("Name", "@NAME")
    ])
    p.add_tools(hover_nodes)

    # Show legend
    p.legend.location = "top_left"

    # Show the plot
    show(p)

def calculate_mode_probabilities(air_utility_df, car_utility_df, rail_utility_df):
    """
    Calculate the choice probabilities for air, road, and rail travel for each OD pair.
    
    Parameters:
    air_utility_df (DataFrame): DataFrame with systematic utilities for air travel for each OD pair.
    car_utility_df (DataFrame): DataFrame with systematic utilities for road travel for each OD pair.
    rail_utility_df (DataFrame): DataFrame with systematic utilities for rail travel for each OD pair.
    
    Returns:
    Tuple[DataFrame, DataFrame, DataFrame]: DataFrames containing the probabilities of choosing air, road, and rail travel respectively.
    """
    # Calculate the exponential of the utilities
    exp_air_utility = np.exp(air_utility_df)
    exp_car_utility = np.exp(car_utility_df)
    exp_rail_utility = np.exp(rail_utility_df)

    # Calculate the sum of exponential utilities for each OD pair
    sum_exp_utilities = exp_air_utility.add(exp_car_utility).add(exp_rail_utility)

    # Calculate the probabilities for each OD pair
    P_air = exp_air_utility.div(sum_exp_utilities)
    P_road = exp_car_utility.div(sum_exp_utilities)
    P_rail = exp_rail_utility.div(sum_exp_utilities)

    # Replace any NaNs that resulted from -inf utilities with zero probability
    P_air = P_air.fillna(0)
    P_road = P_road.fillna(0)
    P_rail = P_rail.fillna(0)

    return P_air, P_road, P_rail

def mirror_od_pairs(od_pairs):
    # Create a copy to avoid modifying the dictionary during iteration
    new_pairs = od_pairs.copy()

    for (origin, destination), paths in od_pairs.items():
        reverse_key = (destination, origin)
        if reverse_key not in od_pairs:  # Check if reverse path already exists to avoid duplication
            reversed_paths = []
            for path in paths:
                reversed_path = {
                    'path': list(reversed(path['path'])),  # Reverse the path
                    'inv_time': path['inv_time'],  # Invariant properties
                    'og_area': destination,  # Switch origin and destination
                    'access_time': path['egress_time'],  # Switch access and egress times
                    'dest_area': origin,
                    'egress_time': path['access_time'],
                    'aggr_total_time': path['aggr_total_time'],
                    'distance': path['distance'],
                    'costs': path['costs'],
                    'transfers': path['transfers'],
                    'utility': path['utility'],
                    'probability': path['probability']  # Probabilities should be the same
                }
                reversed_paths.append(reversed_path)

            new_pairs[reverse_key] = reversed_paths

    # Update the original dictionary to include the mirrored paths
    od_pairs.update(new_pairs)

def check_city_name_correspondence(od_pairs, pass_rail):
    # Extract city names from the dictionary keys
    dict_cities = set()
    for (origin, destination), paths in od_pairs.items():
        dict_cities.update([origin, destination])

    # Extract city names from the DataFrame's index and columns
    dataframe_cities = set(pass_rail.index).union(set(pass_rail.columns))

    # Check and print non-matching cities
    non_matching_in_dict = dict_cities - dataframe_cities
    non_matching_in_dataframe = dataframe_cities - dict_cities

    if non_matching_in_dict:
        print("Cities in the dictionary not found in the DataFrame:", non_matching_in_dict)
    if non_matching_in_dataframe:
        print("Cities in the DataFrame not found in the dictionary:", non_matching_in_dataframe)
    if not non_matching_in_dict and not non_matching_in_dataframe:
        print("The cities in the OD-pair dictionary fit the cities in the passenger OD-matrix.")

def add_passengers_to_paths(od_pairs, pass_rail):
    # Iterate through each OD-pair in the dictionary
    for (origin, destination), paths in od_pairs.items():
        if origin in pass_rail.index and destination in pass_rail.columns:
            # Fetch the number of passengers from the DataFrame
            passengers = pass_rail.at[origin, destination]
            # Update each path in this OD-pair
            for path in paths:
                probability = path['probability']
                # Calculate the number of passengers for this specific path
                path_passengers = passengers * probability
                # Add the passengers count to the path dictionary
                path['passengers'] = path_passengers

def calculate_and_assign_def_passengers(od_pairs, G):
    for od_pair, paths in od_pairs.items():
        for path_details in paths:
            # Get the path for the current path entry
            path = path_details['path']
            
            # Initialise induced demand to 0 and set default path flow
            induced_demand = 0
            path_flow = path_details['passengers']  # default if no induced demand
            
            # Iterate over the path to check each edge
            for i in range(len(path) - 1):
                n1 = path[i]
                n2 = path[i + 1]
                
                # Check if the edge exists in the graph G and check for 'New Construction'
                if G.has_edge(n1, n2):
                    edge_data = G.get_edge_data(n1, n2)
                    if edge_data.get('status') == 'New Construction':
                        induced_demand = 1
                        path_flow = path_details['passengers'] * 1.25  # Apply increased flow if induced demand is triggered
                        break  # Stop checking further as one triggering edge is enough
            
            # Assign the calculated values to the path details
            path_details['Induced Demand'] = induced_demand
            path_details['path_flow'] = path_flow

def update_link_flows(G, od_pairs):
    # Set initial link_flow to 0 for all edges
    nx.set_edge_attributes(G, 0, 'link_flow')

    # Iterate over each OD-pair in the od_pairs dictionary
    for od_pair, paths in od_pairs.items():
        for path_details in paths:
            path = path_details['path']  # Extract the path
            path_flow = path_details['path_flow']  # Extract the path flow

            # Iterate over the edges in the path and update link_flow
            for i in range(len(path) - 1):
                # Create the edge tuple (n1, n2)
                n1, n2 = path[i], path[i + 1]

                # Check if the edge exists in the graph G and update link_flow
                if G.has_edge(n1, n2):
                    # If the edge exists, add the path_flow to the link_flow
                    if 'link_flow' in G[n1][n2]:
                        G[n1][n2]['link_flow'] += path_flow
                    else:
                        G[n1][n2]['link_flow'] = path_flow
                else:
                    # Handle the case where the edge might be missing if needed
                    print(f"Warning: Edge ({n1}, {n2}) not found in the graph.")

def update_station_node_attributes(od_pairs, G):
    # Initialise node attributes for passenger flow
    nx.set_node_attributes(G, 0, 'origin/termin')
    nx.set_node_attributes(G, 0, 'traverse')
    
    # Dictionary to hold the sum of originating and terminating passengers for each node
    origin_termin_sum = {}

    for od_pair, paths in od_pairs.items():
        for path_details in paths:
            path_flow = path_details['path_flow']
            path = path_details['path']

            # Update origin and termination counters
            for node in [path[0], path[-1]]:  # Only look at the first and last node for origin/termination
                if G.nodes[node].get('is_station', False):  # Check if it is a station node
                    origin_termin_sum[node] = origin_termin_sum.get(node, 0) + path_flow
            
            # Update traverse counter for intermediate nodes
            for node in path[1:-1]:  # Traverse the intermediate nodes
                if G.nodes[node].get('is_station', False):  # Check if it is a station node
                    G.nodes[node]['traverse'] = G.nodes[node].get('traverse', 0) + path_flow

    # Update the 'origin/termin' attribute of the nodes in the graph
    for node, flow_sum in origin_termin_sum.items():
        G.nodes[node]['origin/termin'] = flow_sum

def plot_graph_passenger_distribution(G):
    # Ensure data is present
    if not nx.get_edge_attributes(G, 'link_flow'):
        raise ValueError("No 'link_flow' data found on edges")

    # Define figure
    p = figure(height=600, width=950,
               x_axis_type="mercator", y_axis_type="mercator")

    # Add the CARTODBPOSITRON tile source for the map background
    p.add_tile(CARTODBPOSITRON)

    # Create a color mapper for 'link_flow' values
    all_link_flows = [data['link_flow'] for _, _, data in G.edges(data=True)]
    all_link_flows = [max(flow, 1) for flow in all_link_flows]  # Ensure all values are > 0 for LogColorMapper
    reversed_palette = Viridis256[::-1]
    color_mapper = LogColorMapper(palette=reversed_palette, low=min(all_link_flows), high=max(all_link_flows))

    # Convert node positions to Web Mercator and prepare for plotting
    pos = {node: lat_lon_to_web_mercator(data['lat'], data['lon']) for node, data in G.nodes(data=True)}
    
    # Plot edges and assign colors based on 'link_flow'
    edge_xs, edge_ys, colors, weights, speeds, statuses = [], [], [], [], [], []
    for edge in G.edges(data=True):
        node_start, node_end, edge_data = edge
        x_start, y_start = pos[node_start]
        x_end, y_end = pos[node_end]
        edge_xs.append([x_start, x_end])
        edge_ys.append([y_start, y_end])
        colors.append(edge_data['link_flow'])
        weights.append(edge_data.get('weight', 'N/A'))  # Default to 'N/A' if not available
        speeds.append(edge_data.get('speed', 'N/A'))
        statuses.append(edge_data.get('status', 'N/A'))

    # Create a ColumnDataSource with edge attributes
    source_edges = ColumnDataSource({
        'xs': edge_xs,
        'ys': edge_ys,
        'link_flow': colors,
        'weight': weights,
        'speed': speeds,
        'status': statuses
    })

    # Add the edges to the figure
    p.multi_line('xs', 'ys', source=source_edges,
                 line_color={'field': 'link_flow', 'transform': color_mapper},
                 line_width=3, line_alpha=0.6)

    # Add color bar at the right of the plot
    color_bar = ColorBar(color_mapper=color_mapper, label_standoff=12, border_line_color=None, location=(0,0))
    p.add_layout(color_bar, 'right')

    # Prepare node data for plotting and hover information
    station_xs, station_ys, lats, lons, areas, fids, acc_times, names, og_ter_pas, trav_pas = [], [], [], [], [], [], [], [], [], []
    for node, (x, y) in pos.items():
        data = G.nodes[node]
        if data.get('is_station', False):  # Check if node is a station
            station_xs.append(x)
            station_ys.append(y)
            lats.append(data['lat'])
            lons.append(data['lon'])
            areas.append(data['AREA'])
            fids.append(data['fid'])
            acc_times.append(data['ACC_TIME'])
            names.append(data['NAME'])
            og_ter_pas.append(data['origin/termin'])
            trav_pas.append(data['traverse'])
	

    # Create a ColumnDataSource for station nodes
    source_stations = ColumnDataSource({
        'x': station_xs,
        'y': station_ys,
        'lat': lats,
        'lon': lons,
        'AREA': areas,
        'fid': fids,
        'ACC_TIME': acc_times,
        'NAME': names,
        'origin_termin': og_ter_pas,
        'traverse': trav_pas
    })

    # Plot station nodes
    p.circle('x', 'y', source=source_stations, size=10, color="#fc8d59", fill_alpha=0.8)

    # Add separate hover tools for edges and nodes
    hover_edges = HoverTool(renderers=[p.multi_line('xs', 'ys', source=source_edges)], tooltips=[
        ("Link Flow", "@link_flow"),
        ("Weight", "@weight"),
        ("Speed", "@speed"),
        ("Status", "@status")
    ])

    hover_nodes = HoverTool(renderers=[p.circle('x', 'y', source=source_stations)], tooltips=[
        ("Latitude", "@lat"),
        ("Longitude", "@lon"),
        ("Area", "@AREA"),
        ("FID", "@fid"),
        ("Access Time", "@ACC_TIME"),
        ("Name:", "@NAME"),
        ("origin_termin", "@origin_termin"),
        ("traverse", "@traverse")
    ])

    # Add hover tools to the plot
    p.add_tools(hover_edges)
    p.add_tools(hover_nodes)

    # Show the plot
    show(p)

def add_total_p_attribute(graph):
    for node, attrs in graph.nodes(data=True):
        # Calculate the sum of 'origin/termin' and 'traverse'
        origin_termin = attrs.get('origin/termin', 0)  # Default to 0 if not found
        traverse = attrs.get('traverse', 0)            # Default to 0 if not found
        total_p = origin_termin + traverse
        
        # Set the new attribute 'total_p'
        graph.nodes[node]['total_p'] = total_p